@@ -1,36 +1,37 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
+from decimal import Decimal
+
 from trytond.model import fields
 from trytond.pyson import Eval, Bool
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 
+from trytond.modules.product import price_digits
+
 __all__ = ['ShipmentOut']
-__metaclass__ = PoolMeta
 
 
-class ShipmentOut:
+class ShipmentOut(metaclass=PoolMeta):
     __name__ = 'stock.shipment.out'
     carrier = fields.Many2One('carrier', 'Carrier', states={
-            'readonly': Eval('state') != 'draft',
+            'readonly': ~Eval('state').in_(['draft', 'waiting', 'assigned',
+                    'packed']),
             },
         depends=['state'])
     cost_currency = fields.Many2One('currency.currency',
-            'Cost Currency', states={
+        'Cost Currency', states={
             'invisible': ~Eval('carrier'),
             'required': Bool(Eval('carrier')),
             'readonly': ~Eval('state').in_(['draft', 'waiting', 'assigned',
                     'packed']),
             }, depends=['carrier', 'state'])
-    cost_currency_digits = fields.Function(fields.Integer(
-        'Cost Currency Digits'),
-        'on_change_with_cost_currency_digits')
     cost = fields.Numeric('Cost',
-            digits=(16, Eval('cost_currency_digits', 2)), states={
+        digits=price_digits, states={
             'invisible': ~Eval('carrier'),
             'readonly': ~Eval('state').in_(['draft', 'waiting', 'assigned',
                     'packed']),
-            }, depends=['carrier', 'state', 'cost_currency_digits'])
+            }, depends=['carrier', 'state'])
     cost_invoice_line = fields.Many2One('account.invoice.line',
             'Cost Invoice Line', readonly=True)
 
@@ -42,39 +43,30 @@ class ShipmentOut:
                     'product "%s".'),
                 })
 
-    @fields.depends('cost_currency')
-    def on_change_with_cost_currency_digits(self, name=None):
-        if self.cost_currency:
-            return self.cost_currency.digits
-        return 2
-
     def _get_carrier_context(self):
         return {}
 
     def get_carrier_context(self):
         return self._get_carrier_context()
 
-    @fields.depends(methods=['inventory_moves'])
+    @fields.depends(methods=['on_change_inventory_moves'])
     def on_change_carrier(self):
-        return self.on_change_inventory_moves()
+        self.on_change_inventory_moves()
 
-    @fields.depends('carrier', 'customer', 'inventory_moves')
+    @fields.depends('carrier', 'customer', 'inventory_moves',
+        methods=['_get_carrier_context'])
     def on_change_inventory_moves(self):
-        Currency = Pool().get('currency.currency')
-
         try:
-            result = super(ShipmentOut, self).on_change_inventory_moves()
+            super(ShipmentOut, self).on_change_inventory_moves()
         except AttributeError:
-            result = {}
+            pass
         if not self.carrier:
-            return result
+            return
         with Transaction().set_context(self._get_carrier_context()):
             cost, currency_id = self.carrier.get_sale_price()
-        currency = Currency(currency_id)
-        result['cost'] = cost
-        result['cost_currency'] = currency.id
-        result['cost_currency_digits'] = currency.digits if currency else 2
-        return result
+        self.cost = cost.quantize(
+            Decimal(1) / 10 ** self.__class__.cost.digits[1])
+        self.cost_currency = currency_id
 
     def _get_cost_tax_rule_pattern(self):
         'Get tax rule pattern for invoice line'
@@ -88,10 +80,10 @@ class ShipmentOut:
 
         if not self.cost:
             return {}
-        with Transaction().set_user(0, set_context=True):
-            invoice_line = InvoiceLine()
+        invoice_line = InvoiceLine()
         product = self.carrier.carrier_product
         invoice_line.type = 'line'
+        invoice_line.product = product
 
         party = invoice.party
         party_context = {}
@@ -106,18 +98,23 @@ class ShipmentOut:
         if invoice.currency != self.cost_currency:
             with Transaction().set_context(date=invoice.currency_date):
                 cost = Currency.compute(self.cost_currency, cost,
-                    invoice.currency)
-        invoice_line.unit_price = cost
+                    invoice.currency, round=False)
+        invoice_line.unit_price = cost.quantize(
+            Decimal(1) / 10 ** InvoiceLine.unit_price.digits[1])
 
         taxes = []
         pattern = self._get_cost_tax_rule_pattern()
         for tax in product.customer_taxes_used:
             if party.customer_tax_rule:
-                tax_ids = party.customer_tax_rule.apply(None, pattern)
+                tax_ids = party.customer_tax_rule.apply(tax, pattern)
                 if tax_ids:
                     taxes.extend(tax_ids)
                 continue
             taxes.append(tax.id)
+        if party.customer_tax_rule:
+            tax_ids = party.customer_tax_rule.apply(None, pattern)
+            if tax_ids:
+                taxes.extend(tax_ids)
         invoice_line.taxes = taxes
 
         invoice_line.account = product.account_revenue_used
